/*
 *                   jfjji,.           H A M M E R H E A D  S H A R K W A R E
 *                  ,#######EL:                  http://www.hhsw.de
 *                   .,j######Kj       
 *                    ;W########G:     This  code  is  free software; you can
 *                   j#####G ,D#L:     redistribute it and/or modify it under
 *             .,;;,G#####D.           the  terms  of  the GNU General Public
 *           jE##########G             License   as  published  by  the  Free
 *             .,W######D.             Software  Foundation; either version 2
 *             :L#####K##G             of  the  License,  or (at your option)
 *           .f#####G;.t##:            any later version.
 *         ;L###ELi.    ,D,            
 *       iEEGWj                        This  code  is distributed in the hope
 *      tW;                            that  it  will  be useful, but without
 *     jf                              any warranty; without even the implied
 *    .Kt                              warranty of merchantability or fitness
 *    i:                               for  a particular purpose. See the GNU
 *    ;                                General   Public   License   for  more
 *    .                                details (http://www.gnu.org).
 *
 * This software is based on bbbutton by Konstantinos Mantzaris
 * <GODoal@hotmail.com> with ideas from wmNetscapeKiller by Pierre-Olivier 
 * <pblumo@free.fr>.
 *
 * @author mf@hhsw.de
 * @version 20041014
 */
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <string>
 
#include "bbdock.hh"

using namespace std;

/*
 * Initialize object
 *
 * @param pngfile - png image file
 * @param command - system command to execute
 * @param fw - frame width in pixels (optional)
 * @param fh - frame height in pixels (optional)
 * @param pfd - writeable pipe to the invoking process (optional)
 */
BBDock::BBDock( char *pngfile, char *command, int fw, int fh, int pfd )
{
	// ensure command will detach from this process by
	// adding an ampersand if there is none
	int	len;

	if( !(len = strlen( command )) )
		cmd = NULL;
	else
	{
		for( cmd = command+len; 
			strchr( " \t\r\n", *(--cmd) ); 
			*cmd = 0 );

		if( *cmd != '&' )
			len += 2; // blank and ampersand

		if( (cmd = new char[++len]) )
			snprintf( cmd, len, "%s &", command );
	}

	// load icon
	loadPNG( pngfile );

	if( !width || !height )
	{
		cerr << "Invalid image file" << endl;
		exit( -1 );
	}

	if(	!(display = XOpenDisplay( NULL )) )
	{
		cerr << "Unable to open display" << endl;
		exit( -1 );
	}

	screen = DefaultScreen( display );
	gc = XDefaultGC( display, screen );

	Window	root;
	Window	dummy;

	root = RootWindow( display, screen );

	{
		dummy = XCreateSimpleWindow( display, root, 
			0, 0, 1, 1, 1, 0, 0 );

		XSizeHints	sh;

	    sh.x = 0;
		sh.y = 0;
		sh.width = fw;
		sh.height = fh;

	    XSetWindowAttributes	xswat;
		unsigned long			vmask;
	
		xswat.background_pixmap = ParentRelative;
		vmask = CWBackPixmap;

		window = XCreateWindow( display, root, 
			sh.x, sh.y, 
			sh.width, sh.height,
			0, 
			CopyFromParent, 
			InputOutput,
			CopyFromParent, 
			vmask, &xswat );
			
		sh.flags = USSize | USPosition;
		XSetWMNormalHints( display, dummy, &sh );

		{
			XWMHints	xwmh;
    
			xwmh.initial_state = WithdrawnState;
			xwmh.icon_window = window;
			xwmh.icon_x = sh.x;
			xwmh.icon_y = sh.y;
			xwmh.window_group = dummy;
			xwmh.flags = StateHint | IconWindowHint |
				IconPositionHint | WindowGroupHint;

			XSetWMHints( display, dummy, &xwmh );
		}
		
		if( (left = (sh.width-width)>>1) < 0 ||
			(top = (sh.height-height)>>1) < 0 )
		{
			cerr << "Image (" << width << "x" <<
				height << ") is too big !" << endl;
			exit( -1 );
		}
	}

	{
		XClassHint	xch;

		xch.res_name = "bbdock";
		xch.res_class = "bbdock";

		XSetClassHint( display, dummy, &xch );
	}

	// determine bits per pixel of an screen-compatible XImage, does not need
	// to be the same as the color-depth of the screen
	{
		XPixmapFormatValues	*pf;
		int					formats;
		
		bitsperpixel = DefaultDepth( display, screen );
		pf = XListPixmapFormats( display, &formats );

		if( formats )
		{
			int	format;
			
			for( format = formats;
			     format--; )
				if( pf[format].depth == bitsperpixel )
				{
					bitsperpixel = pf[format].bits_per_pixel;
					break; 
				}

			XFree( (char *) pf );
		}
	}

	{
		int	bytesperline;

		// ceil to full byte if bits per pixel is fewer than a byte
		// or divide by 8 if there are at last 8 bits per pixel
		if( bitsperpixel < 8 )
			bytesperline = ((width*bitsperpixel)+7) & ~7;
		else
			bytesperline = (width*bitsperpixel)>>3;
						
		// pad every image line to a multiple of 4 bytes
		padding = bytesperline;
		bytesperline = (bytesperline+3) & ~3;
		padding = bytesperline-padding;

		if( !(bitmap = (char *) calloc( 
				bytesperline*height, sizeof( char ) )) ||
			!(surface = XCreateImage( display, 
				DefaultVisual( display, screen ), 
				DefaultDepth( display, screen ), 
				ZPixmap, 0, bitmap, 
				width, height, 
				32, 0 )) )
		{
			cerr << "Could not create XImage" << endl;
			exit( -1 );
		}
	}

	XSelectInput( display, window,
		ButtonReleaseMask | ButtonPressMask |
		ExposureMask | VisibilityChangeMask | 
		StructureNotifyMask | SubstructureNotifyMask | 
		PropertyChangeMask );

	XMapWindow( display, dummy );

	// signal invoking process to continue execution,
	// this signal transfer guarantees the icons will
	// appear in the right order
	if( pfd )
	{
		char	signal = Loaded;

		write( pfd, &signal, sizeof( signal ) );
		close( pfd );
	}
}

/*
 * Clean up
 */
BBDock::~BBDock()
{
	// memory allocated for surface will be freed 
	// by X11
	if( surface )
		XDestroyImage( surface );

	if( cmd )
		delete cmd;
}

/*
 * Load PNG icon, it shouldn't be much greater than 48x48 pixels
 * because this function will not resize anything
 *
 * @param filename - PNG file to load
 */
void BBDock::loadPNG( char *filename )
{
	png_structp	png = NULL;
	png_infop	info = NULL;
	FILE		*fp = NULL;

	width = height = 0;

	if( !(png = png_create_read_struct( PNG_LIBPNG_VER_STRING,
		NULL, NULL, NULL )) )
		return;
 
 	for( ;; )
	{
		if( !(info = png_create_info_struct( png )) ||
			setjmp( png_jmpbuf( png ) ) ||
			!(fp = fopen( filename, "rb" )) )
			break;

		png_init_io( png, fp );
		png_set_sig_bytes( png, 0 );
		png_read_png( png, info, 
			PNG_TRANSFORM_IDENTITY | 
			PNG_TRANSFORM_EXPAND |
			PNG_TRANSFORM_BGR,
			NULL );

		if( !(info->valid & PNG_INFO_IDAT) ||
			!(icon = new int[info->width*info->height]) )
			break;

		for( int y = 0, *src, *dest = icon; 
			y < info->height && (src = (int *) info->row_pointers[y]); 
			y++, dest += info->width )
			memcpy( dest, src, info->width<<2 );

		width = info->width;
		height = info->height;
		size = width*height;

		break;
	}

	if( fp )
		fclose( fp );

	png_destroy_read_struct( &png, &info, (png_infopp) NULL );
}

/*
 * Draw icon to surface, this method is a little slow and one of
 * some reasons why this software is version 0.1
 */
void BBDock::drawIcon()
{
	unsigned char	*dest = (unsigned char *) bitmap;
	unsigned char	*src = (unsigned char *) icon;

	switch( bitsperpixel )
	{
		default:
		case 1:
		case 2:
		case 4:
		case 8:
			// I'm a little lazy ;-) By the way, at this pixel 
			// depth you probably go better with bbbutton
			cerr << "Pixel depth not (yet) supported" << endl;
			exit( -1 );
			break;
		case 16:
			// read below
			src--;
			for( int y = height; y--; dest += padding )
				for( int x = width; x--; )
					switch( *(src+4) )
					{
						case 0:
							src += 4;
							dest += 2;
							break;
						case 0xff:
							// increment must be in front or g++
							// (at least to version 3.3) wont do it
							// correctly, first bit is unused
							*((short *) dest) = 
								((short) ((*(++src)&0xf8)>>3)) |
								((short) ((*(++src)&0xf8)<<3)) |
								((short) ((*(++src)&0xf8)<<8));
							src++;
							dest += 2;
							break;
						default:
							{
								double	diff;
								double	mod = (double)
									255/(*(src+4));
									
								short	pixel = *((short *) dest);

								int	blue = (pixel<<3)&0xf8;
								int green = (pixel>>3)&0xf8;
								int red = (pixel>>8)&0xf8;
								
								diff = *(++src);
								diff -= blue;
								diff /= mod;
								blue += (int) diff;
								
								diff = *(++src);
								diff -= green;
								diff /= mod;
								green += (int) diff;

								diff = *(++src);
								diff -= red;
								diff /= mod;
								red += (int) diff;

								*((short *) dest) = 
									((short) ((blue&0xf8)>>3)) |
									((short) ((green&0xf8)<<3)) |
									((short) ((red&0xf8)<<8));

								dest += 2;
								src++;
							}
							break;
					}
			break;
		case 24:
			for( int y = height; y--; dest += padding )
				for( int x = width; x--; )
					switch( *((int *) src) & 0xff000000 )
					{
						case 0:
							src += 4;
							dest += 3;
							break;
						case 0xff000000:
							*(dest++) = *(src++);
							*(dest++) = *(src++);
							*(dest++) = *(src++);
							src++;
							break;
						default:
							{
								double	diff;
								double	mod = (double)
									255/(*(src+3));

								diff = *(src++);
								diff -= *dest;
								diff /= mod;
								*(dest++) += (int) diff;

								diff = *(src++);
								diff -= *dest;
								diff /= mod;
								*(dest++) += (int) diff;

								diff = *(src++);
								diff -= *dest;
								diff /= mod;
								*(dest++) += (int) diff;

								src++;
							}
							break;
					}
			break;
		case 32:
			for( int p = size; p--; )
				switch( *((int *) src) & 0xff000000 )
				{
					case 0:
						src += 4;
						dest += 4;
						break;
					case 0xff000000:
						*((int *) dest) = *((int *) src);
						src += 4;
						dest += 4;
						break;
					default:
						{
							double	diff;
							double	mod = (double)
								255/(*(src+3));

							diff = *(src++);
							diff -= *dest;
							diff /= mod;
							*(dest++) += (int) diff;

							diff = *(src++);
							diff -= *dest;
							diff /= mod;
							*(dest++) += (int) diff;

							diff = *(src++);
							diff -= *dest;
							diff /= mod;
							*(dest++) += (int) diff;

							src++;
							dest++;
						}
						break;
				}
			break;
	}
}

/*
 * Intercept BadMatch that can occur when the window
 * is withdrawn, there is no opportunity to quest for this
 * as far as I can see
 */
int throwError( Display *, XErrorEvent * )
{
	return 0;
}

/*
 * Draw icon
 *
 * @param displacement - vertical and horizontal displacement
 */
void BBDock::paint( int displacement )
{
	/* doesn't work !
	XWindowAttributes	xattr;
 	
	XGetWindowAttributes( display, window, &xattr );

	if( xattr.map_state != IsViewable )
		return;*/

	XClearWindow( display, window );

	// this is a quick and ugly hack to avoid a BadMatch error
	// that occurs in the following situation: slit is in auto-hide
	// mode and hidden, open it and right click on its border to
	// open the slit menu, now right-click somewhere on the desktop 
	// and brace yourself ;-) Unfortunately the check above doesn't 
	// work for this kind of situation and it's far too late now for
	// a better solution so this one is left for version 0.2
	XErrorHandler	handler = XSetErrorHandler( throwError );

	XGetSubImage( display, window, left, top, width, height, 
		0xffffffff, ZPixmap, surface, 0, 0 );

	XSetErrorHandler( handler );

	drawIcon();

	XPutImage( display, window, gc, surface, 0, 0, 
		left+displacement, top+displacement, 
		width, height );
}
			
/*
 * X11 event loop
 */
void BBDock::run()
{
	for( bool exposed = false; ; )
	{
		XEvent	event;
		
		bzero( &event, sizeof( event ) );
		XNextEvent( display, &event );

		switch( event.type )
		{
			case Expose:
				if( !event.xexpose.count )
				{
					exposed = true;
					paint();
				}
				break;
			case ConfigureNotify:
				// for updates in size and position
				if( exposed &&
					event.xconfigure.window == window )
					paint();
				break;
			case ButtonPress:
				switch( event.xbutton.button )
				{
					case 1:
						paint( 1 );
						break;
				}
				break;
			case ButtonRelease:
				switch( event.xbutton.button )
				{
					case 1:
						paint();
						system( cmd );
						break;
				}
				break;
		}
    }
}
